<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE kpartgui SYSTEM "kpartgui.dtd">
<data name="effects" version="0">
  <group list="invert,sepia,tcolor,greyscale,frei0r.B,frei0r.G,frei0r.R,frei0r.contrast0r,frei0r.saturat0r,frei0r.tint0r,frei0r.primaries,chroma_hold,frei0r.colorize,frei0r.equaliz0r,frei0r.hueshift0r,frei0r.luminance,lumaliftgaingamma,lift_gamma_gain,brightness,gamma,frei0r.colgate,frei0r.balanc0r,frei0r.brightness,frei0r.levels,frei0r.three_point_balance,frei0r.curves,frei0r.coloradj_RGB,frei0r.sopsat,frei0r.bezier_curves,avfilter.selectivecolor,avfilter.lut3d">
    <text>Colour</text>
  </group>
  <group list="loudness,dynamic_loudness,volume,gain,ladspa.1049,ladspa.1048,ladspa.1413,mute,fadein,fadeout,audiobalance,audiopan,swapchannels,ladspa.1406,channelcopy">
    <text>Audio correction</text>
  </group>
  <group list="audiowave,audiospectrum,frei0r.pr0be,frei0r.pr0file,frei0r.rgbparade,frei0r.timeout,frei0r.vectorscope">
    <text>Analysis and data</text>
  </group>
  <group list="fade_from_black,fade_to_black,boxblur,obscure,autotrack_rectangle,frei0r.squareblur,frei0r.glow,frei0r.IIRblur,frei0r.edgeglow,frei0r.medians,frei0r.softglow,burningtv,charcoal,dust,frei0r.cartoon,frei0r.dither,frei0r.emboss,frei0r.posterize,frei0r.rgbnoise,frei0r.sigmoidaltransfer,frei0r.sobel,frei0r.threelay0r,frei0r.threshold0r,frei0r.twolay0r,frei0r.vignette,grain,lines,lightgraffiti,oldfilm,threshold,wave,mirror,frei0r.distort0r,frei0r.lenscorrection,frei0r.defish0r,frei0r.pixeliz0r,frei0r.d90stairsteppingfix,frei0r.hqdn3d,frei0r.sharpness,frei0r.spillsupress">
    <text>Image adjustment</text>
  </group>
  <group list="freeze">
    <text>Motion</text>
  </group>
  <group list="chroma,frei0r.alpha0ps,frei0r.alphagrad,frei0r.alphaspot,frei0r.transparency,frei0r.select0r,frei0r.mask0mate,rotoscoping,frei0r.keyspillm0pup,frei0r.c0rners,crop,affine,qtblend,vignette,frei0r.letterb0xed,pan_zoom,frei0r.scale0tilt,affinerotate,frei0r.nosync0r">
    <text>Alpha/Transform</text>
  </group>
  <group list="movit.blur,movit.sharpen,movit.diffusion,movit.glow,movit.lift_gamma_gain,movit.mirror,movit.opacity,movit.rect,movit.saturation,movit.unsharp_mask,movit.vignette,movit.white_balance">
    <text>GPU effects</text>
  </group>
</data>
